from typing import Optional, Union, List

from pydantic import BaseModel

from nidhoggr.utils.transformer import YggdrasilRequestTransformer, JSONResponseTransformer


class UserProperty(BaseModel):
    name: str
    value: str

    class Config:
        allow_mutation = False
        ignore_extra = False

    def __hash__(self):
        return hash((self.name, self.value))

    def verify(self, *, key: bytes):
        return True

    @property
    def unsigned(self):
        return self


class SignedUserProperty(UserProperty):
    signature: Optional[str] = None

    class Config:
        allow_mutation = False
        ignore_extra = False

    def __hash__(self):
        return hash((self.name, self.value, self.signature))

    def verify(self, *, key: bytes):
        from nidhoggr.utils.crypto import verify_property
        return verify_property(public_key=key, prop=self)

    @property
    def unsigned(self):
        return UserProperty(name=self.name, value=self.value)


class JoinRequest(BaseModel, YggdrasilRequestTransformer):
    accessToken: str
    selectedProfile: str
    serverId: str

    class Config:
        allow_mutation = False


class HasJoinedRequest(BaseModel, YggdrasilRequestTransformer):
    username: str
    serverId: str
    ip: str

    class Config:
        allow_mutation = False


class JoinedResponse(BaseModel, JSONResponseTransformer):
    id: str
    name: str
    properties: Union[List[UserProperty], List[SignedUserProperty]]

    class Config:
        allow_mutation = False


class ProfileRequest(BaseModel, YggdrasilRequestTransformer):
    id: str
    unsigned: bool = False

    class Config:
        allow_mutation = False



from functools import partial

import pytest
from flask.testing import FlaskClient

from nidhoggr.views import session
from ..conftest import accessor, EndpointCallable


@pytest.fixture
def join(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(session.join), client)


@pytest.fixture
def has_joined(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(session.has_joined), client)


@pytest.fixture
def profile(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(session.profile), client)



from functools import partial

import pytest
from flask.testing import FlaskClient

from nidhoggr.views import auth
from ..conftest import accessor, EndpointCallable


@pytest.fixture
def authenticate(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(auth.authenticate), client)


@pytest.fixture
def refresh(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(auth.refresh), client)


@pytest.fixture
def validate(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(auth.validate), client)


@pytest.fixture
def invalidate(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(auth.invalidate), client)


@pytest.fixture
def signout(client: FlaskClient) -> EndpointCallable:
    return partial(accessor(auth.signout), client)


def check_uuid(uuid: str):
    assert isinstance(uuid, str)
    assert len(uuid) == 32
    assert isinstance(int(uuid, 16), int)



from nidhoggr.errors.auth import InvalidAccessToken, InvalidClientToken
from nidhoggr.errors.common import BadPayload
from ..conftest import cast


def test_empty_credentials(user, validate):
    assert cast[BadPayload](validate({}))
    assert cast[BadPayload](validate({"clientToken": user.client}))


def test_broken_access_token(user, validate):
    response = validate({
        "accessToken": "anything",
        "clientToken": user.client
    })
    assert response.status_code == 403
    assert cast[InvalidAccessToken](response)


def test_invalid_client_token(user, validate):
    response = validate({
        "accessToken": user.access,
        "clientToken": "anything"
    })
    assert response.status_code == 403
    assert cast[InvalidClientToken](response)


def test_validate_full(user, validate):
    response = validate({
        "accessToken": user.access,
        "clientToken": user.client
    })
    assert response.status_code == 204
    assert response.data == b""


def test_validate_simple(user, validate):
    response = validate({
        "accessToken": user.access,
    })
    assert response.status_code == 204
    assert response.data == b""



"""Errors for Yggdrasil protocol"""

from nidhoggr.errors.common import YggdrasilError

MigrationDone = YggdrasilError(
    status=403,
    errorMessage="Use e-mail instead of plain login",
    cause="UserMigratedException",
)
InvalidCredentials = YggdrasilError(
    status=403,
    error="ForbiddenOperationException",
    errorMessage="Invalid credentials. Invalid username, email or password",
)
InvalidAccessToken = YggdrasilError(
    status=403,
    error="ForbiddenOperationException",
    errorMessage="accessToken expired or not exists",
)
InvalidClientToken = YggdrasilError(
    status=403,
    error="ForbiddenOperationException",
    errorMessage="Collision with clientToken, or it's broken",
)
InvalidProfile = YggdrasilError(
    status=400,
    errorMessage="Can't use selected profile",
)



import json
from functools import wraps
from itertools import product
from pathlib import Path
from typing import Optional, Dict, Any, Union, Callable, TypeVar

import pytest
from flask import url_for, Response, Flask

from nidhoggr.application import create_app
from nidhoggr.bl import BaseUserRepo, User, BLConfig, BaseTextureRepo
from nidhoggr.errors.common import YggdrasilError
from nidhoggr.models.auth import AuthenticationResponse, RefreshResponse
from nidhoggr.models.session import JoinedResponse
from nidhoggr.models.texture import TextureType, TextureItem, TextureResponse, TextureMeta, TextureToken
from nidhoggr.utils.crypto import KeyPair


class TestUser(User):
    # NOTE: Still needed for testing purposes
    password: str


USERS = list(map(
    TestUser.parse_obj,
    json.loads((Path(__file__).parent / "data/users.json").read_text())
))

EndpointCallable = Callable[[Dict[str, Any]], Response]
TEST_KEY_PAIR = KeyPair.generate()

YggdrasilResponse = Union[
    AuthenticationResponse,
    RefreshResponse,
    JoinedResponse,
    TextureResponse,
]


def accessor(func):
    @wraps(func)
    def wrapper(client, data):
        return client.post(
            url_for(func.__name__),
            data=json.dumps(data),
            content_type='application/json'
        )

    return wrapper


T = TypeVar("T")


class cast:
    def __getitem__(self, other) -> Callable[[T], Union[YggdrasilError, YggdrasilResponse]]:
        if isinstance(other, YggdrasilError):
            return lambda res: self[YggdrasilError](res) == other
        elif issubclass(other, YggdrasilError):
            return lambda res: other(status=res.status_code, **res.json)
        elif other in YggdrasilResponse.__args__:
            return lambda res: other(**res.json)
        else:
            return lambda res: None


cast = cast()


class TestUserRepo(BaseUserRepo):
    __users: Dict[int, TestUser]

    def __init__(self, *, users):
        self.__users = {user.uuid: user.copy() for user in users}

    def get_user(self, **kw: Dict[str, str]) -> Optional[TestUser]:
        res = [
            u
            for u
            in self.__users.values()
            if any(getattr(u, k) == v for k, v in kw.items())
        ]
        return (res or [None])[0]

    def save_user(self, *, user: TestUser):
        self.__users[user.uuid] = user

    def check_password(self, *, clean: str, uuid: str) -> bool:
        user = self.get_user(uuid=uuid)
        return user and user.password == clean


class TestTextureRepo(BaseTextureRepo):
    def get(self, *, uuid: str) -> Dict[TextureType, TextureItem]:
        return {}

    def save(self, uuid: str, texture_type: TextureType, payload: bytes, metadata: TextureMeta) -> bool:
        pass

    def delete(self, *, uuid: str, texture_type: TextureType):
        pass

    def fetch_remote(self, *, url: str) -> bytes:
        pass

    def save_token(self, *, name: str, ip: str, server_id: str) -> TextureToken:
        pass

    def get_token(self, *, value: str) -> TextureToken:
        pass


@pytest.fixture(params=list(product((False, True), (False, True))))
def config(request) -> BLConfig:
    strict, simple_login, *_ = request.param
    return BLConfig(
        key_pair=TEST_KEY_PAIR,
        strict=strict,
        simple_login=simple_login,
    )


@pytest.fixture
def users() -> TestUserRepo:
    return TestUserRepo(users=USERS)


@pytest.fixture
def textures() -> TestTextureRepo:
    return TestTextureRepo()


@pytest.fixture(autouse=True)
def skip_bl(request, config):
    if request.node.get_marker("skip_bl_on"):
        marker = request.node.get_marker("skip_bl_on")
        if any(
            marker.kwargs.get(name, ...) == value
            for name, value
            in config._asdict().items()
        ):
            pytest.skip(f"Skipped on {config}")


@pytest.fixture
def user(users) -> User:
    return users.get_user(uuid="1")


@pytest.fixture
def old_user(users) -> User:
    return users.get_user(uuid="2")


@pytest.fixture
def new_user(users) -> User:
    return users.get_user(uuid="3")


@pytest.fixture
def app(
    users,
    config: BLConfig,
    textures: BaseTextureRepo
) -> Flask:
    return create_app(
        users=users,
        config=config,
        textures=textures,
    )



from datetime import datetime
from typing import Union

from nidhoggr.bl import BaseUserRepo, BLConfig, BaseTextureRepo
from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import YggdrasilError
from nidhoggr.errors.session import InvalidServer
from nidhoggr.models.auth import EmptyResponse, NoContent
from nidhoggr.models.session import JoinRequest, HasJoinedRequest, JoinedResponse, ProfileRequest
from nidhoggr.models.texture import TextureResponse
from nidhoggr.utils.crypto import sign_property
from nidhoggr.utils.decorator import typed


@typed
def join(req: JoinRequest, users: BaseUserRepo) -> Union[EmptyResponse, YggdrasilError]:
    user = users.get_user(access=req.accessToken)

    if user is None:
        return InvalidProfile

    if req.selectedProfile != user.uuid:
        return InvalidProfile

    user = user.copy(update=dict(server=req.serverId))
    users.save_user(user=user)
    return NoContent


@typed
def has_joined(req: HasJoinedRequest, users: BaseUserRepo) -> Union[JoinedResponse, YggdrasilError]:
    user = users.get_user(login=req.username)

    if user is None:
        return InvalidProfile

    if req.serverId != user.server:
        return InvalidServer

    return JoinedResponse(
        id=user.uuid,
        name=user.login,
        properties=user.properties
    )


@typed
def profile(
    req: ProfileRequest,
    users: BaseUserRepo,
    config: BLConfig,
    textures: BaseTextureRepo
) -> Union[JoinedResponse, YggdrasilError]:
    user = users.get_user(uuid=req.id)

    if user is None:
        return InvalidProfile

    textures = textures.get(uuid=req.id)

    texture_prop = TextureResponse(
        timestamp=int(datetime.now().timestamp() * 1000),
        profileId=user.uuid,
        profileName=user.login,
        textures=textures
    ).pack()

    raw_props = [texture_prop, *user.properties]

    if req.unsigned:
        properties = raw_props
    else:
        properties = [
            sign_property(private_key=config.key_pair.private, prop=prop)
            for prop
            in raw_props
        ]

    return JoinedResponse(
        id=user.uuid,
        name=user.login,
        properties=properties
    )


__all__ = {"join", "has_joined", "profile"}



from typing import Optional

from pydantic import BaseModel

from nidhoggr.utils.transformer import JSONErrorTransformer


class YggdrasilError(BaseModel, JSONErrorTransformer):
    """Base error class for Yggdrasil-specific errors."""

    status: int
    errorMessage: str
    error: str = "YggdrasilError"
    cause: Optional[str] = None


BadPayload = YggdrasilError(
    status=400,
    errorMessage="Malformed request: incorrect arguments",
)
UnknownError = YggdrasilError(
    status=500,
    errorMessage="Unknown internal error",
)
MethodNotAllowed = YggdrasilError(
    status=405,
    errorMessage=(
        "The method specified in the request is not allowed "
        "for the resource identified by the request URI. "
        "Yggdrasil support only POST requests. "
    ),
)
NotFound = YggdrasilError(
    status=404,
    errorMessage="The server has not found anything matching the request URI",
)
BadRequest = YggdrasilError(
    status=400,
    errorMessage=(
        "The server is refusing to service the request "
        "because the entity of the request is in a format "
        "not supported by the requested resource for the requested method"
    ),
)



from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import BadPayload
from nidhoggr.errors.session import InvalidServer
from nidhoggr.models.session import JoinedResponse
from nidhoggr.utils.crypto import generate_uuid
from ..conftest import cast


def test_empty_credentials(user, has_joined):
    assert cast[BadPayload](has_joined({}))
    assert cast[BadPayload](has_joined({"username": user.login}))
    assert cast[BadPayload](has_joined({"ip": "127.0.0.1"}))
    assert cast[BadPayload](has_joined({"serverId": generate_uuid()}))
    assert cast[BadPayload](has_joined({"username": user.login, "ip": "127.0.0.1"}))
    assert cast[BadPayload](has_joined({"username": user.login, "serverId": generate_uuid()}))
    assert cast[BadPayload](has_joined({"ip": "127.0.0.1", "serverId": generate_uuid()}))


def test_invalid_server(user, join, has_joined):
    join({
        "accessToken": user.access,
        "selectedProfile": user.uuid,
        "serverId": generate_uuid()
    })
    has_joined_response = has_joined({
        "username": user.login,
        "serverId": "anything",
        "ip": "127.0.0.1",
    })

    assert cast[InvalidServer](has_joined_response)


def test_nonexistent_user(has_joined):
    response = has_joined({
        "username": "anything",
        "serverId": "anything",
        "ip": "127.0.0.1",
    })

    assert cast[InvalidProfile](response)


def test_has_joined(user, join, has_joined):
    server_id = generate_uuid()
    join({
        "accessToken": user.access,
        "selectedProfile": user.uuid,
        "serverId": server_id
    })
    has_joined_response = has_joined({
        "username": user.login,
        "serverId": server_id,
        "ip": "127.0.0.2",
    })
    result = cast[JoinedResponse](has_joined_response)
    assert result is not None
    assert result.id == user.uuid
    assert result.name == user.login



from abc import ABCMeta, abstractmethod
from typing import Optional, List, Dict, NamedTuple

from pydantic import BaseModel

from nidhoggr.models.session import UserProperty
from nidhoggr.models.texture import TextureType, TextureItem, TextureMeta, TextureToken
from nidhoggr.utils.crypto import KeyPair


class User(BaseModel):
    uuid: str
    login: str
    email: str
    access: Optional[str] = None
    client: Optional[str] = None
    server: Optional[str] = None
    properties: List[UserProperty] = []

    class Config:
        allow_mutation = False


class BLConfig(NamedTuple):
    key_pair: KeyPair
    strict: bool = False
    simple_login: bool = True
    optimize_textures: bool = True


class BaseUserRepo(metaclass=ABCMeta):

    @abstractmethod
    def get_user(self, **kw: Dict[str, str]) -> Optional[User]:
        pass

    @abstractmethod
    def check_password(self, *, clean: str, uuid: str) -> bool:
        pass

    @abstractmethod
    def save_user(self, *, user: User) -> bool:
        pass


class BaseTextureRepo(metaclass=ABCMeta):
    @abstractmethod
    def get(self, *, uuid: str) -> Dict[TextureType, TextureItem]:
        pass

    @abstractmethod
    def save(self, uuid: str, texture_type: TextureType, payload: bytes, metadata: TextureMeta) -> bool:
        pass

    @abstractmethod
    def delete(self, *, uuid: str, texture_type: TextureType):
        pass

    @abstractmethod
    def fetch_remote(self, *, url: str) -> bytes:
        pass

    @abstractmethod
    def save_token(self, *, name: str, ip: str, server_id: str) -> TextureToken:
        pass

    @abstractmethod
    def get_token(self, *, value: str) -> TextureToken:
        pass



from typing import Union

from nidhoggr.bl import BaseUserRepo, BLConfig
from nidhoggr.errors.auth import MigrationDone, InvalidCredentials, InvalidAccessToken, InvalidClientToken
from nidhoggr.errors.common import YggdrasilError, BadRequest
from nidhoggr.models import auth
from nidhoggr.utils.crypto import generate_uuid
from nidhoggr.utils.decorator import typed


@typed
def authenticate(
    req: auth.AuthenticationRequest,
    users: BaseUserRepo,
    config: BLConfig
) -> Union[auth.AuthenticationResponse, YggdrasilError]:
    user = users.get_user(email=req.username, login=req.username)

    if user is None:
        return InvalidCredentials

    if not users.check_password(clean=req.password, uuid=user.uuid):
        return InvalidCredentials

    if not config.simple_login and req.username == user.login:
        return MigrationDone

    if (config.strict and (user.client is None)) or (not config.strict):
        client_token = req.clientToken or generate_uuid()
        user = user.copy(update=dict(client=client_token))
    elif config.strict and req.clientToken is not None and user.client != req.clientToken:
        return InvalidClientToken

    user = user.copy(update=dict(access=generate_uuid()))

    if req.agent is not None:
        profile = auth.Profile(id=user.uuid, name=user.login)
        selected_profile, available_profiles = profile, [profile]
    else:
        selected_profile = available_profiles = None

    if req.requestUser:
        user_instance = auth.UserInstance(id=user.uuid, properties=user.properties)
    else:
        user_instance = None

    users.save_user(user=user)

    return auth.AuthenticationResponse(
        accessToken=user.access,
        clientToken=user.client,
        availableProfiles=available_profiles,
        selectedProfile=selected_profile,
        user=user_instance,
    )


@typed
def refresh(req: auth.RefreshRequest, users: BaseUserRepo) -> Union[auth.RefreshResponse, YggdrasilError]:
    user = users.get_user(client=req.clientToken)

    if user is None:
        return BadRequest

    user = user.copy(update=dict(access=generate_uuid()))
    selected_profile = auth.Profile(id=user.uuid, name=user.login)

    if req.requestUser:
        user_instance = auth.UserInstance(id=user.uuid, properties=user.properties)
    else:
        user_instance = None

    users.save_user(user=user)

    return auth.RefreshResponse(
        accessToken=user.access,
        clientToken=user.client,
        selectedProfile=selected_profile,
        user=user_instance,
    )


@typed
def validate(req: auth.ValidateRequest, users: BaseUserRepo) -> Union[auth.EmptyResponse, YggdrasilError]:
    user = users.get_user(access=req.accessToken)

    if user is None:
        return InvalidAccessToken

    if req.clientToken is not None and req.clientToken != user.client:
        return InvalidClientToken

    return auth.NoContent


@typed
def invalidate(req: auth.InvalidateRequest, users: BaseUserRepo) -> Union[auth.EmptyResponse, YggdrasilError]:
    user = users.get_user(access=req.accessToken)

    if user is None:
        return InvalidAccessToken

    if req.clientToken != user.client:
        return InvalidClientToken

    user = user.copy(update=dict(access=None))

    users.save_user(user=user)

    return auth.NoContent


@typed
def signout(req: auth.SignOutRequest, users: BaseUserRepo) -> Union[auth.EmptyResponse, YggdrasilError]:
    user = users.get_user(email=req.username)

    if user is None:
        return InvalidCredentials

    if not users.check_password(clean=req.password, uuid=user.uuid):
        return InvalidCredentials

    user = user.copy(update=dict(access=None))

    users.save_user(user=user)

    return auth.NoContent


__all__ = {"authenticate", "refresh", "validate", "invalidate", "signout"}



import pytest
from flask import url_for, Response
from flask.testing import FlaskClient

from nidhoggr.utils.crypto import generate_uuid
from nidhoggr.views import auth
from .conftest import check_uuid


def test_generate_uuid():
    check_uuid(generate_uuid())


def test_not_found(client: FlaskClient):
    response: Response = client.get("/unknown")
    assert response.status_code == 404


@pytest.mark.parametrize(["endpoint"], [(f,) for f in auth.__all__])
def test_get_method(client: FlaskClient, endpoint: str):
    response: Response = client.get(url_for(endpoint))
    assert response.status_code == 405



import logging
from functools import wraps
from typing import Callable

from flask import current_app, Response
from pydantic import ValidationError

from nidhoggr.errors.common import BadPayload
from nidhoggr.utils.transformer import AbstractRequestTransformer, AbstractResponseTransformer


logger = logging.getLogger()


def _transform_request(cls, *view_args, **view_kwargs):
    if issubclass(cls, AbstractRequestTransformer):
        return cls.transform(*view_args, **view_kwargs)
    else:
        return current_app.bl.get(cls)


def as_response(func):
    @wraps(func)
    def wrapper(error):
        response: AbstractResponseTransformer = func(error)
        return response.transform()

    return wrapper


def typed(func: Callable):
    @wraps(func)
    def wrapper(*view_args, **view_kwargs) -> Response:
        try:
            # noinspection PyUnresolvedReferences
            args = [
                _transform_request(arg_cls, *view_args, **view_kwargs)
                for arg_name, arg_cls
                in func.__annotations__.items()
                if arg_name != "return"
            ]
        except ValidationError:
            response: AbstractResponseTransformer = BadPayload
        else:
            response: AbstractResponseTransformer = func(*args)
        return response.transform()

    return wrapper



from abc import ABCMeta, abstractmethod

from flask import request, Response
from pydantic import BaseModel

MIME_TYPE = 'application/json'


class AbstractRequestTransformer(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def transform(cls, *args, **kwargs) -> BaseModel:
        pass


class AbstractResponseTransformer(metaclass=ABCMeta):
    @abstractmethod
    def transform(self, *args, **kwargs) -> Response:
        pass


class EmptyResponseTransformer(AbstractResponseTransformer):
    @classmethod
    def transform(cls, *args, **kwargs):
        return Response(response="", status=204)


class YggdrasilRequestTransformer(AbstractRequestTransformer):
    @classmethod
    def transform(cls: BaseModel, *args, **kwargs):
        return cls.parse_raw(request.data)


class JSONResponseTransformer(AbstractResponseTransformer):
    def transform(self: BaseModel, *args, **kwargs):
        return Response(
            mimetype=MIME_TYPE,
            content_type=MIME_TYPE,
            response=self.json()
        )


class JSONErrorTransformer(AbstractResponseTransformer):
    def transform(self: BaseModel, *args, **kwargs):
        return Response(
            status=self.status,
            mimetype=MIME_TYPE,
            content_type=MIME_TYPE,
            response=self.copy(exclude={'status'}).json()
        )



from pydantic import BaseModel

from nidhoggr.utils.transformer import JSONErrorTransformer


class TextureError(BaseModel, JSONErrorTransformer):
    status: int
    message: str


EmptySkinType = TextureError(message="Requested type is empty", status=400)

StorageError = TextureError(message="Can't save texture", status=500)

UnknownUser = TextureError(message="Unknown user", status=404)

UnsupportedFormat = TextureError(message="Unsupported format", status=400)

UnsupportedSize = TextureError(message="Unsupported size", status=400)

MissingToken = TextureError(message="Authorization token not provided or it has expired", status=401)

InvalidToken = TextureError(message="Invalid authorization token", status=403)

InvalidName = TextureError(message="Wrong name", status=403)

InvalidIP = TextureError(message="IP does not match", status=403)

NotJoined = TextureError(message="User has not joined yet", status=403)



import pytest

from nidhoggr.bl import BaseUserRepo
from nidhoggr.errors.auth import MigrationDone, InvalidCredentials, InvalidClientToken
from nidhoggr.errors.common import BadPayload
from nidhoggr.models.auth import Agent, UserInstance, AuthenticationResponse
from nidhoggr.utils.crypto import generate_uuid
from .conftest import check_uuid
from ..conftest import cast

nonstring_values = ([1], {2: 3}, None)


def test_empty_credentials(authenticate, user):
    assert cast[BadPayload](authenticate({}))
    assert cast[BadPayload](authenticate({"username": user.email}))
    assert cast[BadPayload](authenticate({"password": user.password}))


@pytest.mark.parametrize(["password"], [(v,) for v in nonstring_values])
@pytest.mark.parametrize(["username"], [(v,) for v in nonstring_values])
def test_nonstring_credentials(authenticate, username, password):
    response = authenticate({
        "username": username,
        "password": password
    })
    assert cast[BadPayload](response)


def test_nonexistent_user(authenticate):
    response = authenticate({
        "username": "unknown@ponyville.eq",
        "password": "anything"
    })
    assert response.status_code == 403
    assert cast[InvalidCredentials](response)


def test_wrong_password(user, authenticate):
    response = authenticate({
        "username": user.email,
        "password": "anything"
    })
    assert response.status_code == 403
    assert cast[InvalidCredentials](response)


@pytest.mark.skip_bl_on(simple_login=True)
def test_migration_done(user, authenticate):
    response = authenticate({
        "username": user.login,
        "password": user.password
    })
    assert response.status_code == 403
    assert cast[MigrationDone](response)


@pytest.mark.skip_bl_on(simple_login=False)
def test_success_simple_login(user, authenticate):
    response = authenticate({
        "username": user.login,
        "password": user.password
    })
    assert response.status_code == 200


def test_success_simple(user, authenticate):
    response = authenticate({
        "username": user.email,
        "password": user.password
    })
    assert response.status_code == 200


def test_access_token_refreshed(users: BaseUserRepo, user, authenticate):
    response = authenticate({
        "username": user.email,
        "password": user.password
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    assert result.accessToken != user.access
    fresh = users.get_user(uuid=user.uuid)
    assert fresh.access == result.accessToken


def test_access_token_generated(users: BaseUserRepo, old_user, authenticate):
    response = authenticate({
        "username": old_user.email,
        "password": old_user.password
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    check_uuid(result.accessToken)
    fresh = users.get_user(uuid=old_user.uuid)
    assert fresh.access == result.accessToken


def test_client_token_generated(users: BaseUserRepo, new_user, authenticate):
    response = authenticate({
        "username": new_user.email,
        "password": new_user.password,
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    check_uuid(result.clientToken)
    fresh = users.get_user(uuid=new_user.uuid)
    assert fresh.client == result.clientToken


def test_client_token_mirrored_back(user, authenticate):
    response = authenticate({
        "username": user.email,
        "password": user.password,
        "clientToken": user.client
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    check_uuid(result.clientToken)
    assert user.client == result.clientToken


@pytest.mark.skip_bl_on(strict=True)
def test_client_token_rewritten(users: BaseUserRepo, user, authenticate):
    new_client_token = generate_uuid()
    response = authenticate({
        "username": user.email,
        "password": user.password,
        "clientToken": new_client_token
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    check_uuid(result.clientToken)
    fresh = users.get_user(uuid=user.uuid)
    assert new_client_token == result.clientToken
    assert fresh.client == result.clientToken


@pytest.mark.skip_bl_on(strict=False)
def test_client_token_rejected(user, authenticate):
    new_client_token = generate_uuid()
    response = authenticate({
        "username": user.email,
        "password": user.password,
        "clientToken": new_client_token
    })
    assert cast[InvalidClientToken](response)


def test_profiles_request(user, authenticate):
    response = authenticate({
        "username": user.email,
        "password": user.password,
        "agent": Agent().dict()
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    assert isinstance(result.availableProfiles, list)
    assert len(result.availableProfiles) == 1
    assert result.availableProfiles[0] == result.selectedProfile


def test_user_instance_request(user, authenticate):
    response = authenticate({
        "username": user.email,
        "password": user.password,
        "requestUser": True
    })
    assert response.status_code == 200
    result = cast[AuthenticationResponse](response)
    assert result is not None
    assert isinstance(result.user, UserInstance)



from nidhoggr.errors.common import UnknownError, MethodNotAllowed, NotFound
from nidhoggr.utils.decorator import as_response


@as_response
def not_found(_):
    return NotFound


@as_response
def method_not_allowed(_):
    return MethodNotAllowed


@as_response
def internal_server_error(error):
    return UnknownError.copy(update=dict(cause=f"{error.__class__.__name__}: {error}"))



from base64 import b64encode
from enum import Enum
from typing import Optional, Dict

from flask import request
from pydantic import BaseModel

from nidhoggr.models.session import UserProperty
from nidhoggr.utils.transformer import JSONResponseTransformer, JSONErrorTransformer, AbstractRequestTransformer

TextureMeta = Optional[Dict[str, str]]


class TextureType(Enum):
    SKIN = "SKIN"
    CAPE = "CAPE"
    ELYTRA = "ELYTRA"


class TextureItem(BaseModel):
    url: str
    metadata: TextureMeta = None

    class Config:
        allow_mutation = False


class TextureGetRequest(BaseModel, AbstractRequestTransformer):
    uuid: str

    class Config:
        allow_mutation = False

    @classmethod
    def transform(cls, *args, **kwargs):
        return cls(uuid=kwargs.get("uuid"))


class TextureSetRequest(BaseModel, AbstractRequestTransformer):
    uuid: str
    texture_type: TextureType
    payload: bytes
    metadata: TextureMeta = None
    access: Optional[str] = None

    class Config:
        allow_mutation = False

    @classmethod
    def transform(cls, *args, **kwargs):
        return cls(
            uuid=kwargs.get("uuid"),
            texture_type=kwargs.get("texture_type").upper(),
            payload=request.files["file"],
            metadata={
                name: value
                for name, value
                in request.form.items()
                if name != "file"
            },
            access=request.headers.get("Authorization")
        )


class TextureDeleteRequest(BaseModel, AbstractRequestTransformer):
    uuid: str
    texture_type: TextureType
    access: Optional[str] = None

    class Config:
        allow_mutation = False

    @classmethod
    def transform(cls, *args, **kwargs):
        return cls(
            uuid=kwargs.get("uuid"),
            texture_type=kwargs.get("texture_type").upper(),
            access=request.headers.get("Authorization")
        )


class TextureResponse(BaseModel, JSONResponseTransformer):
    timestamp: int
    profileId: str
    profileName: str
    textures: Dict[TextureType, TextureItem]

    class Config:
        allow_mutation = False
        use_enum_values = True

    def pack(self) -> UserProperty:
        return UserProperty(
            name="textures",
            value=b64encode(self.json().encode('ascii')).decode('ascii')
        )


class TextureToken(BaseModel):
    name: str
    ip: str
    server_id: str
    value: str

    class Config:
        allow_mutation = False


class TextureHandshakeFirstRequest(BaseModel, AbstractRequestTransformer):
    name: str
    ip: str

    class Config:
        allow_mutation = False

    @classmethod
    def transform(cls, *args, **kwargs):
        return cls(
            name=request.form.get("name"),
            ip=request.remote_addr,
        )


class TextureHandshakeFirstResponse(BaseModel, JSONErrorTransformer):
    offline: bool = False
    serverId: str
    verifyToken: str

    class Config:
        allow_mutation = False


class TextureHandshakeSecondRequest(BaseModel, AbstractRequestTransformer):
    name: str
    verifyToken: str

    class Config:
        allow_mutation = False

    @classmethod
    def transform(cls, *args, **kwargs):
        return cls(
            name=request.form.get("name"),
            ip=request.form.get("verifyToken"),
        )


class TextureHandshakeSecondResponse(BaseModel, JSONErrorTransformer):
    accessToken: str
    userId: str

    class Config:
        allow_mutation = False


class TextureStatusResponse(BaseModel, JSONErrorTransformer):
    message: str


TextureSuccess = TextureStatusResponse(message="OK")



from base64 import b64encode, b64decode
from typing import NamedTuple
from uuid import uuid4

from M2Crypto import EVP, RSA, BIO

from nidhoggr.models.session import UserProperty, SignedUserProperty


class KeyPair(NamedTuple):
    private: bytes
    public: bytes

    @classmethod
    def generate(cls, *, bits: int = 4096) -> "KeyPair":
        """
        Generate an RSA keypair with an exponent of 65537 in PEM format

        :param: bits The key length in bits
        :return: Key pair with private and public keys
        """
        key = RSA.gen_key(bits, 65537)

        with BIO.MemoryBuffer() as buffer:
            key.save_key_bio(buffer, cipher=None)
            private_key = buffer.getvalue()

        with BIO.MemoryBuffer() as buffer:
            key.save_pub_key_bio(buffer)
            public_key = buffer.getvalue()

        del key
        return KeyPair(private=private_key, public=public_key)


def sign_property(*, private_key: bytes, prop: UserProperty) -> SignedUserProperty:
    key = EVP.load_key_string(private_key)
    key.reset_context(md="sha1")
    key.sign_init()
    key.sign_update(prop.value.encode("ascii"))
    signature = b64encode(key.sign_final()).decode("ascii")
    return SignedUserProperty(name=prop.name, value=prop.value, signature=signature)


def verify_property(*, public_key: bytes, prop: SignedUserProperty) -> bool:
    with BIO.MemoryBuffer(public_key) as buffer:
        key = EVP.load_key_bio_pubkey(buffer)
    key.reset_context(md="sha1")
    key.verify_init()
    key.verify_update(prop.value.encode("ascii"))
    result = (key.verify_final(b64decode(prop.signature)) == 1)
    del key
    return result


def generate_uuid() -> str:
    """Generate random UUID token like Java's UUID.toString()"""
    return uuid4().hex



from nidhoggr.bl import BaseUserRepo
from nidhoggr.errors.auth import InvalidAccessToken, InvalidClientToken
from nidhoggr.errors.common import BadPayload
from ..conftest import cast


def test_empty_credentials(user, invalidate):
    assert cast[BadPayload](invalidate({}))
    assert cast[BadPayload](invalidate({"accessToken": user.access}))
    assert cast[BadPayload](invalidate({"clientToken": user.client}))
    assert cast[BadPayload](invalidate({"clientToken": user.client}))


def test_broken_access_token(user, invalidate):
    response = invalidate({
        "accessToken": "anything",
        "clientToken": user.client
    })
    assert response.status_code == 403
    assert cast[InvalidAccessToken](response)


def test_invalid_client_token(user, invalidate):
    response = invalidate({
        "accessToken": user.access,
        "clientToken": "anything"
    })
    assert response.status_code == 403
    assert cast[InvalidClientToken](response)


def test_invalidate_full(users: BaseUserRepo, user, invalidate):
    response = invalidate({
        "accessToken": user.access,
        "clientToken": user.client
    })
    assert response.status_code == 204
    assert response.data == b""
    fresh = users.get_user(uuid=user.uuid)
    assert fresh.access is None



from typing import List, Optional

from pydantic import BaseModel

from nidhoggr.models.session import UserProperty
from nidhoggr.utils.transformer import YggdrasilRequestTransformer, JSONResponseTransformer, EmptyResponseTransformer


class Agent(BaseModel):
    name: str = 'Minecraft'
    version: int = 1

    class Config:
        allow_mutation = False


class Profile(BaseModel):
    id: str
    name: str
    legacy: bool = False

    class Config:
        allow_mutation = False


class UserInstance(BaseModel):
    id: str
    properties: List[UserProperty] = []

    class Config:
        allow_mutation = False


class AuthenticationRequest(BaseModel, YggdrasilRequestTransformer):
    agent: Optional[Agent] = None
    username: str
    password: str
    clientToken: Optional[str] = None
    requestUser: bool = False

    class Config:
        allow_mutation = False


class AuthenticationResponse(BaseModel, JSONResponseTransformer):
    accessToken: str
    clientToken: str
    availableProfiles: Optional[List[Profile]] = None
    selectedProfile: Optional[Profile] = None
    user: Optional[UserInstance]

    class Config:
        allow_mutation = False


class RefreshRequest(BaseModel, YggdrasilRequestTransformer):
    accessToken: str
    clientToken: str
    selectedProfile: Optional[Profile] = None
    requestUser: bool = False

    class Config:
        allow_mutation = False


class RefreshResponse(BaseModel, JSONResponseTransformer):
    accessToken: str
    clientToken: str
    selectedProfile: Profile = None
    user: Optional[UserInstance] = None

    class Config:
        allow_mutation = False


class InvalidateRequest(BaseModel, YggdrasilRequestTransformer):
    accessToken: str
    clientToken: str

    class Config:
        allow_mutation = False


class SignOutRequest(BaseModel, YggdrasilRequestTransformer):
    username: str
    password: str

    class Config:
        allow_mutation = False


class ValidateRequest(BaseModel, YggdrasilRequestTransformer):
    accessToken: str
    clientToken: Optional[str] = None

    class Config:
        allow_mutation = False


class EmptyResponse(EmptyResponseTransformer):
    pass


NoContent = EmptyResponse()



import pytest

from nidhoggr.bl import User, BLConfig
from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import BadPayload
from nidhoggr.models.session import UserProperty, SignedUserProperty, JoinedResponse
from ..conftest import cast

with_sign = pytest.mark.parametrize(
    ["unsigned"],
    [(True,), (False,)],
    ids=["unsigned", "signed"]
)


def test_empty_credentials(profile):
    assert cast[BadPayload](profile({}))


@with_sign
def test_nonexistent_user(unsigned: bool, profile):
    assert cast[InvalidProfile](profile({"id": "anything", "unsigned": unsigned}))


@with_sign
def test_profile_simple(user: User, unsigned: bool, profile):
    response = profile({"id": user.uuid, "unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    assert result.id == user.uuid
    assert result.name == user.login


@with_sign
def test_properties(config: BLConfig, new_user: User, unsigned: bool, profile):
    response = profile({"id": new_user.uuid})
    result = cast[JoinedResponse](response)
    assert result is not None
    assert all(
        type(p) is unsigned and UserProperty or SignedUserProperty
        for p
        in result.properties
    )
    assert all(p.verify(key=config.key_pair.public) for p in result.properties)
    assert set(p.unsigned for p in result.properties) >= set(new_user.properties)


@with_sign
def test_texture_exists(user: User, unsigned: bool, profile):
    response = profile({"id": user.uuid, "unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    texture_prop = ([p for p in result.properties if p.name == "textures"] or [None])[0]
    assert isinstance(texture_prop, UserProperty)


@with_sign
def test_texture_attached(user: User, unsigned: bool, profile):
    response = profile({"id": user.uuid, "unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    diff = set(p.unsigned for p in result.properties) - set(user.properties)
    assert len(diff) == 1
    assert diff.pop().name == "textures"



from nidhoggr.bl import BaseUserRepo
from nidhoggr.errors.common import BadPayload, BadRequest
from nidhoggr.models.auth import Profile, UserInstance, RefreshResponse
from ..conftest import cast


def test_empty_credentials(user, refresh):
    assert cast[BadPayload](refresh({}))
    assert cast[BadPayload](refresh({"accessToken": user.access}))
    assert cast[BadPayload](refresh({"clientToken": user.client}))


def test_nonexistent_token_pair(refresh):
    response = refresh({
        "accessToken": "anything",
        "clientToken": "anything"
    })
    assert response.status_code == 400
    assert cast[BadRequest](response)


def test_access_token_refreshed(users: BaseUserRepo, user, refresh):
    response = refresh({
        "accessToken": user.access,
        "clientToken": user.client
    })
    assert response.status_code == 200
    result = cast[RefreshResponse](response)
    assert result is not None
    assert result.accessToken != user.access
    fresh = users.get_user(uuid=user.uuid)
    assert fresh.access == result.accessToken


def test_client_token_mirrored_back(user, refresh):
    response = refresh({
        "accessToken": user.access,
        "clientToken": user.client
    })
    assert response.status_code == 200
    result = cast[RefreshResponse](response)
    assert result is not None
    assert result.clientToken == user.client


def test_selected_profile(user, refresh):
    response = refresh({
        "accessToken": user.access,
        "clientToken": user.client
    })
    assert response.status_code == 200
    result = cast[RefreshResponse](response)
    assert result is not None
    assert isinstance(result.selectedProfile, Profile)


def test_user_instance_request(user, refresh):
    response = refresh({
        "accessToken": user.access,
        "clientToken": user.client,
        "requestUser": True
    })
    assert response.status_code == 200
    result = cast[RefreshResponse](response)
    assert result is not None
    assert isinstance(result.user, UserInstance)



from nidhoggr.bl import BLConfig
from nidhoggr.models.session import UserProperty, SignedUserProperty
from nidhoggr.utils.crypto import sign_property, verify_property


def test_crypto(config: BLConfig):
    unsigned_prop = UserProperty(name="twilight", value="sparkle")
    signed_prop = sign_property(private_key=config.key_pair.private, prop=unsigned_prop)
    assert signed_prop is not None
    assert isinstance(signed_prop, SignedUserProperty)
    assert isinstance(signed_prop.signature, str)
    assert verify_property(public_key=config.key_pair.public, prop=signed_prop)



from io import BytesIO
from typing import Optional, Union

from PIL import Image

from nidhoggr.errors.texture import UnsupportedFormat, UnsupportedSize

AVAILABLE_SIZES = frozenset({64, 128, 256, 512, 1024, 2048})


def check_texture(*, data: bytes) -> Optional[Union[UnsupportedFormat, UnsupportedSize]]:
    with Image.open(BytesIO(data)) as image:
        if image.format != "PNG":
            return UnsupportedFormat
        (width, height) = image.size
        valid = width / 2 == height or width == height
        if not valid or width not in AVAILABLE_SIZES:
            return UnsupportedSize


def optimize_texture(*, data: bytes) -> bytes:
    return data



from nidhoggr.bl import BaseUserRepo
from nidhoggr.errors.auth import InvalidCredentials
from nidhoggr.errors.common import BadPayload
from ..conftest import cast


def test_empty_credentials(user, signout):
    assert cast[BadPayload](signout({}))
    assert cast[BadPayload](signout({"username": user.email}))
    assert cast[BadPayload](signout({"password": user.password}))


def test_nonexistent_user(signout):
    response = signout({
        "username": "unknown@ponyville.eq",
        "password": "anything"
    })
    assert response.status_code == 403
    assert cast[InvalidCredentials](response)


def test_wrong_password(user, signout):
    response = signout({
        "username": user.email,
        "password": "anything"
    })
    assert response.status_code == 403
    assert cast[InvalidCredentials](response)


def test_signout_full(users: BaseUserRepo, user, signout):
    response = signout({
        "username": user.email,
        "password": user.password,
    })
    assert response.status_code == 204
    assert response.data == b""
    fresh = users.get_user(uuid=user.uuid)
    assert fresh.access is None



from nidhoggr.bl import BaseUserRepo
from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import BadPayload
from nidhoggr.utils.crypto import generate_uuid
from ..conftest import cast


def test_empty_credentials(user, join):
    assert cast[BadPayload](join({}))
    assert cast[BadPayload](join({"accessToken": user.access}))
    assert cast[BadPayload](join({"selectedProfile": user.uuid}))
    assert cast[BadPayload](join({"serverId": generate_uuid()}))
    assert cast[BadPayload](join({"accessToken": user.access, "selectedProfile": user.uuid}))
    assert cast[BadPayload](join({"accessToken": user.access, "serverId": generate_uuid()}))
    assert cast[BadPayload](join({"selectedProfile": user.access, "serverId": generate_uuid()}))


def test_invalid_profile(user, join):
    response = join({
        "accessToken": user.access,
        "selectedProfile": "anything",
        "serverId": generate_uuid()
    })
    assert cast[InvalidProfile](response)


def test_nonexistent_user(user, join):
    server_id = generate_uuid()
    response = join({
        "accessToken": "anything",
        "selectedProfile": user.uuid,
        "serverId": server_id
    })

    assert cast[InvalidProfile](response)


def test_join(users: BaseUserRepo, user, join):
    server_id = generate_uuid()
    response = join({
        "accessToken": user.access,
        "selectedProfile": user.uuid,
        "serverId": server_id
    })
    assert response.status_code == 204
    assert response.data == b""
    fresh = users.get_user(uuid=user.uuid)
    assert fresh.server == server_id



from itertools import chain

from flask import Flask
from werkzeug.exceptions import NotFound, MethodNotAllowed, InternalServerError

from nidhoggr.bl import BaseUserRepo, BLConfig, BaseTextureRepo
from nidhoggr.views import auth, session, core


def configure_error_handlers(app: Flask):
    app.register_error_handler(NotFound, core.not_found)
    app.register_error_handler(MethodNotAllowed, core.method_not_allowed)
    app.register_error_handler(InternalServerError, core.internal_server_error)


def configure_views(app: Flask):
    for func in auth.__all__:
        app.add_url_rule(f"/{func}", func, getattr(auth, func), methods=["POST"])
    for func in session.__all__:
        app.add_url_rule(f"/{func}", func, getattr(session, func), methods=["POST"])


def create_app(users: BaseUserRepo, config: BLConfig, textures: BaseTextureRepo) -> Flask:
    app = Flask(__package__)
    app.bl = {
        BaseUserRepo: users,
        BLConfig: config,
        BaseTextureRepo: textures
    }
    configure_error_handlers(app)
    configure_views(app)
    return app



from datetime import datetime
from typing import Union

from nidhoggr.bl import BaseUserRepo, BaseTextureRepo, BLConfig
from nidhoggr.errors import texture as err
from nidhoggr.models import texture as tex
from nidhoggr.models.session import HasJoinedRequest, JoinedResponse
from nidhoggr.utils.crypto import generate_uuid
from nidhoggr.utils.decorator import typed
from nidhoggr.utils.texture import check_texture, optimize_texture
from nidhoggr.views import session


@typed
def get(
    req: tex.TextureGetRequest,
    users: BaseUserRepo,
    textures: BaseTextureRepo
) -> Union[tex.TextureResponse, err.TextureError]:
    user = users.get_user(uuid=req.id)

    if user is None:
        return err.UnknownUser

    result = textures.get(uuid=req.id)

    return tex.TextureResponse(
        timestamp=int(datetime.now().timestamp() * 1000),
        profileId=user.uuid,
        profileName=user.login,
        textures=result
    )


@typed
def save(
    req: tex.TextureSetRequest,
    users: BaseUserRepo,
    textures: BaseTextureRepo,
    config: BLConfig
) -> Union[tex.TextureStatusResponse, err.TextureError]:
    user = users.get_user(uuid=req.id)

    if user is None:
        return err.UnknownUser

    if req.access is None:
        return err.MissingToken

    if req.access != user.access:
        return err.InvalidToken

    texture_error = check_texture(data=req.payload)
    if texture_error is not None:
        return texture_error

    if config.optimize_textures:
        payload = optimize_texture(data=req.payload)
    else:
        payload = req.payload

    result = textures.save(
        uuid=req.uuid,
        texture_type=req.texture_type,
        payload=payload,
        metadata=req.metadata
    )

    return result and tex.TextureSuccess or err.StorageError


@typed
def delete(
    req: tex.TextureDeleteRequest,
    users: BaseUserRepo,
    textures: BaseTextureRepo
) -> Union[tex.TextureStatusResponse, err.TextureError]:
    user = users.get_user(uuid=req.id)

    if user is None:
        return err.UnknownUser

    if req.access is None:
        return err.MissingToken

    if req.access != user.access:
        return err.InvalidToken

    result = textures.get(uuid=req.id)

    if req.texture_type not in result:
        return err.EmptySkinType

    textures.delete(uuid=user.uuid, texture_type=req.texture_type)

    return tex.TextureSuccess


@typed
def handshake_first_stage(
    req: tex.TextureHandshakeFirstRequest,
    textures: BaseTextureRepo
) -> tex.TextureHandshakeFirstResponse:
    server_id = generate_uuid()

    generated_token = textures.save_token(
        name=req.name,
        ip=req.ip,
        server_id=server_id
    )

    return tex.TextureHandshakeFirstResponse(
        serverId=generate_uuid(),
        verifyToken=generated_token.value
    )


@typed
def handshake_second_stage(
    req: tex.TextureHandshakeSecondRequest,
    textures: BaseTextureRepo,
    users: BaseUserRepo
) -> Union[tex.TextureHandshakeSecondResponse, err.TextureError]:
    token = textures.get_token(value=req.verifyToken)

    if token is None:
        return err.MissingToken

    if token.name != req.name:
        return err.InvalidName

    if token.ip != req.ip:
        return err.InvalidIP

    join_request = HasJoinedRequest(username=req.name, serverId=token.server_id, ip=token.ip)
    join_result = session.has_joined.__wrapped__(join_request, users)

    if not isinstance(join_result, JoinedResponse):
        return err.NotJoined

    user = users.get_user(uuid=join_result.id)

    return tex.TextureHandshakeSecondResponse(
        accessToken=user.access,
        userId=user.uuid
    )
